/*
 * xray - simple ray tracing application
 * 
 * Copyright 1991 Samuel T. Denton, III
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that my name not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior
 * permission.  I make no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 * 
 * Author:  Samuel T. Denton, III; St. Louis, MO
 */

#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/Xatom.h>
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <memory.h>
#ifdef USE_SHM
#include <sys/ipc.h>
#include <sys/shm.h>
#ifdef _sys_shm_h
void *shmat(); /* stupid SunOS 4.x DOESN'T prototype function */
#endif
#include <X11/extensions/XShm.h>
#endif

#include "xray_icon"
#include "xray_mask"
#include "xray64_icon"
#include "xray64_mask"
#include "xray_cursor"
#include "xray_cmask"
#define DEFAULT_BORDER_SIZE 2

int             Argc;		/* copy of argc */
char          **Argv;		/* copy of argv */
char           *ProgName;	/* name of program being executed */

Display        *dpy = NULL;	/* X server being used */
char           *ClassName;	/* -name option */
int             screen;		/* screen being used */
XSizeHints     *szHints;	/* for sizes and locations */
Visual         *vis;		/* visual to use */
unsigned        dep;		/* depth of this visual */
XStandardColormap *map_info;/* facts about the colormap */
Window          w;			/* rendering window */

void            init_image(), repaint_image(), create_main_window();
void            get_colormap();;

void 
Exit(status)
	int             status;
{
	if (dpy) {
		XCloseDisplay(dpy);
		dpy = NULL;
	}
	(void) exit(status);
}

Usage()
{
	static char    *help_message[] = {
		"where options include:\n",
		"    -display displaystring\tX server to contact\n",
		"    -geometry geomstring\tsize and location of window\n",
		"    -bw pixels\t\t\tborder width in pixels\n",
		"    -bd color\t\t\tborder color\n",
		"    -bg color\t\t\tbackground color\n",
		"\n",
		"Options may be abbreviated to the shortest unambiguous prefix.\n",
		"\n",
		NULL
	},            **cpp;

	fprintf(stderr, "usage:\n\t%s [-options ...]\n\n", ProgName);
	for (cpp = help_message; *cpp; cpp++) {
		fputs(*cpp, stderr);
	}
	Exit(1);
}

main(argc, argv)
	int             argc;
	char          **argv;
{
	int             i;	/* iterator variable */
	char           *displayname = NULL;	/* name of X server to
						 * contact */
	char           *geometry = NULL;	/* size and location of
						 * enlarged */
	char           *border_color = NULL;	/* border color name */
	char           *back_color = NULL;	/* background of enlargement */
	int             border_width = -1;	/* border width in pixels */
	int             count;
	int             sync = 0;	/* synchronous X */

	ProgName = argv[0];
	Argc = argc;
	Argv = argv;

	for (i = 1; i < argc; i++) {
		char           *arg = argv[i];

		if (arg[0] == '-') {	/* process command line switches */
			switch (arg[1]) {
			case 'd':	/* -display displaystring */
				if (++i >= argc)
					Usage();
				displayname = argv[i];
				continue;
			case 'g':	/* -geometry geomstring */
				if (++i >= argc)
					Usage();
				geometry = argv[i];
				continue;
			case 's':	/* -sync */
				sync++;
				continue;
			case 'n':	/* -name string */
				if (++i >= argc)
					Usage();
				ClassName = argv[i];
				continue;
			case 'b':
				switch (arg[2]) {
				case 'w':	/* -bw pixels */
					if (++i >= argc)
						Usage();
					border_width = atoi(argv[i]);
					continue;
				case 'd':	/* -bd color */
					if (++i >= argc)
						Usage();
					border_color = argv[i];
					continue;
				case 'g':	/* -bg color */
					if (++i >= argc)
						Usage();
					back_color = argv[i];
					continue;
				}	/* end switch */
				Usage();
				/* NOTREACHED */
			default:
				Usage();
				/* NOTREACHED */
			}	/* end switch */
		} else {	/* any command line parameters */
			if (arg[0] == '=') /* =geomstring (obsolete) */
				geometry = arg+1;
			else
				Usage();
		}		/* end if else */
	}			/* end for */

	dpy = XOpenDisplay(displayname);
	if (sync)
		XSynchronize(dpy, True);
	if (!dpy) {
		fprintf(stderr, "%s:  unable to open display \"%s\".\n",
			ProgName, XDisplayName(displayname));
		Exit(1);
	}
	/*
	 * get defaults (would be nice to use the resource manager instead)
	 */

	if (!geometry)
		geometry = XGetDefault(dpy, ProgName, "geometry");
	if (!geometry)
		geometry = XGetDefault(dpy, ProgName, "Geometry");

	if (border_width < 0) {
		char           *tmp = XGetDefault(dpy, ProgName, "borderWidth");
		if (!tmp)
			tmp = XGetDefault(dpy, ProgName, "BorderWidth");

		border_width = tmp ? atoi(tmp) : DEFAULT_BORDER_SIZE;
	}
	if (!border_color)
		border_color = XGetDefault(dpy, ProgName, "borderColor");
	if (!border_color)
		border_color = XGetDefault(dpy, ProgName, "BorderColor");

	if (!back_color)
		back_color = XGetDefault(dpy, ProgName, "background");
	if (!back_color)
		back_color = XGetDefault(dpy, ProgName, "Background");

	/*
	 * set up globals
	 */

	screen = DefaultScreen(dpy);

	/*
	 * do the work
	 */

	get_colormap();
	create_main_window(geometry, border_color, border_width, back_color);

	Exit(0);
}

/*****************************************************************************/
void
get_colormap()
{
	int             i, j, nbr_maps, nbr_visuals;
	XVisualInfo     v_template, *v_list;
	XStandardColormap *map_list = NULL;/* facts about the colormap */

	if (XGetRGBColormaps(dpy, RootWindow(dpy, screen),
			     &map_list, &nbr_maps, XA_RGB_BEST_MAP)) {
		/* ICCCM states nbr_maps should equal one */
	} else if (XGetRGBColormaps(dpy, RootWindow(dpy, screen),
			     &map_list, &nbr_maps, XA_RGB_GRAY_MAP)) {
		/* ICCCM states nbr_maps should equal one */
	} else if (XGetRGBColormaps(dpy, RootWindow(dpy, screen),
				 &map_list, &nbr_maps, XA_RGB_DEFAULT_MAP)) {
		/* ICCCM states nbr_maps may exceed one */
		fprintf(stderr, "%s: no BEST or GRAY colormap, using DEFAULT\n", ProgName);
	} else {
		fprintf(stderr, "%s: no colormaps, run xstdcmap\n",
				ProgName);
		Exit(1);
	}

	map_info = NULL; dep = 0;
fprintf(stderr,"nbr_maps=%d\n",nbr_maps);
	for (i=0; i<nbr_maps; i++) {	/* loop in case of multiple DEFAULT maps */
		v_template.visualid = map_list[i].visualid;
fprintf(stderr,"%d: 0x%X\n",i,map_list[i]);
		v_list = XGetVisualInfo(dpy, VisualIDMask, &v_template, &nbr_visuals);
		if (v_list) {
			for (j=0; j<nbr_visuals; j++) {
				register unsigned d=v_list[j].depth;
				if (d > dep) {
					dep = d;
					vis = v_list[j].visual;
					map_info = map_list+i;
				}
			}
			XFree((char*)v_list);
		} else {
			register unsigned d=DefaultDepth(dpy, screen);
			if (d > dep) {
				dep = d;
				vis = CopyFromParent;
				map_info = map_list+i;
			}
		}
	}
	if (!map_info) {
		fprintf(stderr,"%s: unable to find usable colormap",ProgName);
		Exit(1);
	}
}

unsigned long
get_color(color_name)
	char           *color_name;
{
	XColor          nearest, exact;

	if (XAllocNamedColor(dpy, map_info->colormap, color_name, &nearest, &exact))
		return (nearest.pixel);
	return (0);
}

void 
create_main_window(geometry, border_color, border_width, back_color)
	char           *geometry;	/* size and location of enlargement */
	char           *border_color;	/* color to use for border */
	int             border_width;	/* width of border */
	char           *back_color;	/* color name or pixel value */
{
	XTextProperty   windowName, iconName;
	XWMHints       *wmHints;
	XClassHint     *classHint;
	int             geom_mask;	/* for XParseGeometry results */
	int             i, n;	/* work variable */
	unsigned long   valuemask;	/* for window attributes */
	XSetWindowAttributes attr;	/* for creating window */
	Window          root;	/* root window of screen */
	Bool            done;
	Atom            wdw = XInternAtom(dpy, "WM_DELETE_WINDOW", 0);

	root = RootWindow(dpy, screen);
	szHints = XAllocSizeHints();
	wmHints = XAllocWMHints();
	classHint = XAllocClassHint();
	geom_mask = 0;

#define SizeValues (WidthValue | HeightValue)
#define PosValues (XValue | YValue)
#define AlignRight (XValue | XNegative)
#define AlignBottom (YValue | YNegative)

	if (geometry) {
		geom_mask = XParseGeometry(geometry, &szHints->x, &szHints->y,
					   (unsigned int *) &szHints->width, (unsigned int *) &szHints->height);
	}
	if ((geom_mask & PosValues) == PosValues)
		szHints->flags |= USPosition;
	if ((geom_mask & SizeValues) != SizeValues)
		geom_mask &= ~SizeValues;
	if (geom_mask & SizeValues) {
		szHints->flags |= USSize;
	} else {
		szHints->flags |= PSize;
		szHints->width = 64;
		szHints->height = 64;
	}

	/*
	 * Verify the input and compute the appropriate sizes
	 */

	if (szHints->width <= 0 || szHints->height <= 0) {
		fprintf(stderr, "%s:  bad window size %dx%d.\n",
			ProgName, szHints->width, szHints->height);
		Exit(1);
	}
	/*
	 * Now that we know the size, we can adjust the position as
	 * requested.
	 */

	init_image(szHints->width, szHints->height);
	if ((geom_mask & AlignRight) == AlignRight)
		szHints->x = DisplayWidth(dpy, screen) + szHints->x -
			szHints->width - border_width * 2;
	if ((geom_mask & AlignBottom) == AlignBottom)
		szHints->y = DisplayHeight(dpy, screen) + szHints->y -
			szHints->height - border_width * 2;

	/*
	 * Make the window, but don't map it
	 */

	attr.background_pixel = back_color ?
		get_color(back_color) : BlackPixel(dpy, screen);
	attr.border_pixel = border_color ?
		get_color(border_color) : BlackPixel(dpy, screen);
	attr.event_mask = ButtonPressMask | KeyPressMask |
		ExposureMask | StructureNotifyMask;
	valuemask = CWBackPixel | CWBorderPixel | CWEventMask;

	attr.colormap = map_info->colormap;
	valuemask |= CWColormap;

	w = XCreateWindow(dpy, root,
			  szHints->x, szHints->y,
			  szHints->width, szHints->height, border_width,
			  dep, CopyFromParent, vis, valuemask, &attr);

	windowName.encoding = XA_STRING;
	windowName.format = 8;
	windowName.value = (unsigned char *) "Ray Tracing";
	windowName.nitems = strlen(windowName.value);
	iconName.encoding = XA_STRING;
	iconName.format = 8;
	iconName.value = (unsigned char *) "RayTrace";
	iconName.nitems = strlen(iconName.value);
	wmHints->flags |= InputHint;
	wmHints->input = True;
	wmHints->flags |= IconPixmapHint;
	wmHints->icon_pixmap = XCreateBitmapFromData(dpy, root,
			 xray_icon_bits, xray_icon_width, xray_icon_height);
	wmHints->flags |= IconMaskHint;
	wmHints->icon_mask = XCreateBitmapFromData(dpy, root,
			 xray_mask_bits, xray_mask_width, xray_mask_height);
	wmHints->flags |= IconPixmapHint;
	wmHints->icon_pixmap = XCreateBitmapFromData(dpy, root,
			 xray64_icon_bits, xray64_icon_width, xray64_icon_height);
	wmHints->flags |= IconMaskHint;
	wmHints->icon_mask = XCreateBitmapFromData(dpy, root,
			 xray64_mask_bits, xray64_mask_width, xray64_mask_height);
	classHint->res_class = "RayTrace";
	classHint->res_name = ClassName;
	XSetWMProperties(dpy, w, &windowName, &iconName, Argv, Argc,
			 szHints, wmHints, classHint);
	XChangeProperty(dpy, w, XInternAtom(dpy, "WM_PROTOCOLS", 0),
			XA_ATOM, 32, PropModeReplace,
			(unsigned char *) &wdw, 1);

	/*
	 * Map the window and do the work. Button1 press displays pixel
	 * value. q, Q, or ^C mean quit.
	 */

	XMapWindow(dpy, w);

	for (done = False; !done;) {
		XEvent          event;
		XConfigureEvent *ce = (XConfigureEvent *) & event;
		char            keybuffer[10];
		int             pending, len;
		static int      x[] = {0, 2, 2, 0, 3, 3, 1, 1, 3, 1, 1, 3, 0, 0, 2, 2};
		static int      y[] = {0, 2, 0, 2, 3, 1, 3, 1, 0, 2, 0, 2, 3, 1, 3, 1};
		static int      piece = 16;
		static int      exposed = 0;

		XFlush(dpy);
		if ((pending = XEventsQueued(dpy, QueuedAfterReading)) == 0) {
			if (piece) {
				XExposeEvent    expose;
				piece--;
				expose.x = 0;
				expose.y = 0;
				expose.width = szHints->width;
				expose.height = szHints->height;
				BuildImage(szHints->width, szHints->height,
					   x[piece], y[piece], 4);
				if (exposed)
					repaint_image(&expose, w);
			} else {
				pending++;
			}
		}
		while (pending--) {
			static char    *event_name[] = {"0Event", "1Event",
				"KeyPress", "KeyRelease", "ButtonPress", "ButtonRelease", "MotionNotify",
				"EnterNotify", "LeaveNotify", "FocusIn", "FocusOut", "KeymapNotify",
				"Expose", "GraphicsExpose", "NoExpose", "VisibilityNotify",
				"CreateNotify", "DestroyNotify", "UnmapNotify", "MapNotify", "MapRequest",
				"ReparentNotify", "ConfigureNotify", "ConfigureRequest", "GravityNotify",
				"ResizeRequest", "CirculateNotify", "CirculateRequest", "PropertyNotify",
				"SelectionClear", "SelectionRequest", "SelectionNotify", "ColormapNotify",
				"ClientMessage", "MappingNotify", "LASTEvent",
			};
			XNextEvent(dpy, &event);
			switch (event.type) {

			case MappingNotify:
				XRefreshKeyboardMapping((XMappingEvent *) & event);
				break;

			case ClientMessage:
				if (event.xclient.data.l[0] == wdw)
					done = True;
				break;

			case Expose:
				exposed = 1;
				repaint_image(&event, event.xany.window);
				break;

			case ConfigureNotify:
				break;

			case ButtonPress:
				tron();
				fprintf(stderr, "x=%d, y=%d\n", event.xkey.x, event.xkey.y);
				BuildPixel(szHints->width, szHints->height,
					   event.xkey.x, szHints->height - event.xkey.y);
				troff();
				break;

			case KeyPress:
				len = XLookupString(&event.xkey, keybuffer,
					      sizeof keybuffer, NULL, NULL);
				if (len == 1) {
					switch (keybuffer[0]) {
					case 3:	/* ^C */
					case 'q':
					case 'Q':
						done = True;	/* q, Q, or ^C to quit */
						break;
					}
				}	/* else ignore */
				break;

			default:
				break;
			}
		}
	}
	XUnmapWindow(dpy, w);
	return;
}

char           *pixmap;
XImage         *image;
#ifdef USE_SHM
XShmSegmentInfo shminfo;
#endif

void
init_image(width, height)
{
	int             i;
	unsigned long   black = get_color("black");
	unsigned long   white = get_color("white");

#ifdef USE_SHM
	i = XShmQueryExtension(dpy);
	image = XShmCreateImage(dpy, vis, dep, ZPixmap,
				NULL, &shminfo, width, height);
	shminfo.shmid = shmget(IPC_PRIVATE,
				image->bytes_per_line * image->height,
				IPC_CREAT | 0777);
	shminfo.shmaddr = image->data = pixmap = shmat(shminfo.shmid, NULL, 0);
	shminfo.readOnly = True;
	i = XShmAttach(dpy, &shminfo);
	memset(pixmap, black, image->bytes_per_line * height);
#else
	pixmap = (char*)malloc((width + 2) * (height + 2));
	image = XCreateImage(dpy, vis, dep, ZPixmap,
				1, pixmap, width, height, (dep + 7) && ~7, width + 2);
#define PIXMAP(x,y) pixmap[(width+2)*(y)+(x)]
	memset(pixmap, black, (width + 2) * (height + 2));
	for (i = 0; i < width + 2; i++)
		PIXMAP(i, 0) = PIXMAP(i, height + 1) = white;
	for (i = 0; i < height + 2; i++)
		PIXMAP(0, i) = PIXMAP(width + 1, i) = white;
#undef PIXMAP
#endif
}

void
repaint_image(eventp, w)
	XEvent         *eventp;
	Window          w;
{
	XExposeEvent   *xe = &eventp->xexpose;	/* to get repaint section */
	static GC       gc = (GC) 0;

	if (!gc)
		gc = XCreateGC(dpy, w, 0, NULL);

#ifndef BUGGY
	xe->x = 0;
	xe->width = szHints->width;
#endif
#ifdef USE_SHM
	XShmPutImage(dpy, w, gc, image,
		  xe->x, xe->y, xe->x, xe->y, xe->width, xe->height, False);
#else
	XPutImage(dpy, w, gc, image,
		  xe->x + 1, xe->y + 1, xe->x, xe->y, xe->width, xe->height);
#endif

	XFlush(dpy);
	return;
}

#define PIXMAP(x,y) pixmap[image->bytes_per_line*(y)+(x)]

void 
Dither(x, y, c, i)
	int             x, y, c;
	float           i;
{
	float           r, g, b;
	extern int      _trace;
#ifdef D4
	int             d = (x & 1) | ((y & 1) << 1);
	static float    magic[] = {0.125, 0.625, 0.875, 0.375};
#else
	int             d = (x & 3) | ((y & 3) << 2);
#define F(x) (x+0.5)/16.0
	static float    magic[] = {
		F(0), F(10), F(2), F(8),
		F(13), F(7), F(15), F(5),
		F(3), F(9), F(1), F(11),
	F(12), F(6), F(14), F(4)};
#endif
	r = (c & 1) ? i : 0;
	g = (c & 2) ? i : 0;
	b = (c & 4) ? i : 0;
	y = szHints->height - y - 1;
	PIXMAP(x, y) = map_info->base_pixel +
		(long) (magic[d] + r * map_info->red_max) * map_info->red_mult +
		(long) (magic[d] + g * map_info->green_max) * map_info->green_mult +
		(long) (magic[d] + b * map_info->blue_max) * map_info->blue_mult;
	if (_trace) {
		fprintf(stderr, "r=%g, g=%g, b=%g, pixmap[%d,%d]=%u\n",
			r, g, b, x, y, PIXMAP(x, y));
	}
}
